local g = BittensGlobalTables
local u, oldU, oldUVersion = g.MakeTable("BittensUtilities", 3)
if u == nil then
	return
end

local _G = _G
local SlashCmdList = SlashCmdList
local next = next
local pairs = pairs
local select = select
local type = type
local wipe = wipe

-------------------------------------------------------------- Table Management
function u.GetFromTable(outerTable, ...)
	for i = 1, select("#", ...) do
		if outerTable == nil then
			return nil
		end
		local key = select(i, ...)
		outerTable = outerTable[key]
	end
	return outerTable
end

function u.GetOrMakeTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		if outerTable[key] == nil then
			outerTable[key] = {}
		end
		outerTable = outerTable[key]
	end
	return outerTable
end

function u.MagicallyIncrement(table, variable)
	local val = (table[variable] or 0) + 1
	table[variable] = val
	return val
end

function u.SkipOrUpgrade(table, feature, version)
	local vName = feature .. "Version"
	if table[vName] and table[vName] >= version then
		return "Skip"
	else
		table[vName] = version
	end
end

function u.ShallowCopy(source, dest)
	if dest then
		wipe(dest)
	else
		dest = { }
	end
	for k, v in pairs(source) do
		dest[k] = v
	end
	return dest
end

function u.PrintTable(table)
	if not table then
		print("table is nil")
		return
	end
	
	local empty = true
	for k, v in pairs(table) do
		print(k, v)
		empty = false
	end
	
	if empty then
		print("table is empty")
	end
end

---------------------------------------------------------------- Slash Commands
local slashHandlers, oldSlashHandlers, oldSlashHandlersVersion 
	= g.GetOrMakeTable("BittensSlashHandlers", 1)
local commandHandlers = u.GetOrMakeTable(slashHandlers, "CommandHandlers")
local subCommandHandlers = u.GetOrMakeTable(slashHandlers, "SubCommandHandlers")

local function getOrMakeSlashHandler(command)
	local handler = commandHandlers[command]
	if handler == nil then
		handler = function(userInput)
			local subCommand, rest = userInput:match("^(%S*)%s*(.-)$")
			if subCommand == nil 
				or subCommandHandlers[command][subCommand] == nil then
				
				if subCommandHandlers[command]["<no sub-command>"] then
					subCommandHandlers[command]["<no sub-command>"](userInput)
				else
					print("Possible sub-commands:")
					for sub, _ in pairs(subCommandHandlers[command]) do
						print("   ", sub)
					end
				end
			else
				subCommandHandlers[command][subCommand](rest)
			end
		end
		commandHandlers[command] = handler
	end
	return handler
end

-- "command" can actually be a table of commands, if you wish
function u.RegisterSlashCommand(command, subCommand, func)
	local allCommands
	if type(command) == "table" then
		allCommands = command
	else
		allCommands = { command }
	end
	local fullName = "BITTENS_UTILS_" .. allCommands[1]
	
	for i, command in pairs(allCommands) do
		_G["SLASH_" .. fullName .. i] = "/" .. command
	end
	
	SlashCmdList[fullName] = getOrMakeSlashHandler(command)
	
	u.GetOrMakeTable(subCommandHandlers, command)
		[subCommand or "<no sub-command>"] = func
end

------------------------------------------------------------------------ Events
--local eventHandlers, oldEventHandlers, oldEventHandlersVersion 
--	= g.GetOrMakeTable("BittensEventHandlers", 1)
--
--if not eventHandlers.Frame then
--	eventHandlers.Frame = CreateFrame("frame")
--	eventHandlers.Handlers = { }
--	eventHandlers.EventFrame:SetScript("OnEvent", function(self, event, ...)
--		for handler in next(eventHandlers.Handlers) do
--			local f = handler[event]
--			if f then
--				f[event](...)
--			end
--		end
--	end)
--end

function u.RegisterEventHandler(handler, addonName)
	local frame = CreateFrame("frame")
	for event, _ in pairs(handler) do
		if event == "MY_ADDON_LOADED" then
			if not handler.ADDON_LOADED then
				frame:RegisterEvent("ADDON_LOADED")
			end
		else
			frame:RegisterEvent(event)
		end
	end
	frame:SetScript("OnEvent", function(self, event, ...)
		if handler[event] then
			handler[event](...)
		end
		if event == "ADDON_LOADED" 
			and handler.MY_ADDON_LOADED 
			and addonName == ... then
			
			handler.MY_ADDON_LOADED()
		end
	end)
end

---------------------------------------------------------------- Options Panels

-- Add the functions apply() and refresh() to the returned frame.  They, along
-- with widgets in the frame, should use/modify the values in frame.Settings.
-- If you wish changes in widgets to apply immediately, call frame.refresh() 
-- after changing frame.Settings.
--
-- An empty table represents default values.
function u.CreateOptionsPanel(name, template, parent)	
	local frame = CreateFrame("Frame", nil, nil, template)
	frame.name = name
	
	function frame.Initialize(savedSettingsTableName)
		frame.SavedSettings = u.GetOrMakeTable(_G, savedSettingsTableName)
		frame.Settings = u.ShallowCopy(frame.SavedSettings)
		frame.apply()
	end
	
	-- definition = {
	--     Name = string,
	--     Text = string,
	--     Default = boolean,
	-- }
	function frame.AddCheckBox(definition, anchor, xOffset, yOffset)
		local check = CreateFrame(
			"CheckButton", "tmpCheckBox", frame, "OptionsCheckButtonTemplate")
		check:SetPoint(anchor, xOffset, yOffset)
		tmpCheckBoxText:SetText(definition.Text)
		check:SetScript("OnClick", function()
			frame.Settings[definition.Name] = check:GetChecked() == 1
			frame.apply()
		end)
		u.GetOrMakeTable(frame, "CheckBoxes")[definition] = check
	end
	
	function frame.IsCheckBoxSelected(definition)
		local checked = frame.Settings[definition.Name]
		return checked == nil and definition.Default or checked
	end
	
	function frame.default()
		wipe(frame.Settings)
	end
	
	function frame.refresh()
		for definition, checkBox in u.Pairs(frame.CheckBoxes) do
			checkBox:SetChecked(frame.IsCheckBoxSelected(definition))
		end
		u.Call(frame.CustomRefresh)
	end
	
	function frame.okay()
		u.ShallowCopy(frame.Settings, frame.SavedSettings)
		frame.apply()
	end
	
	function frame.cancel()
		u.ShallowCopy(frame.SavedSettings, frame.Settings)
		frame.apply()
	end
	
	InterfaceOptions_AddCategory(frame)
	return frame
end

-------------------------------------------------------------------- nil safety
function u.Call(func, ...)
	if func then
		func(...)
	end
end

function u.Pairs(table)
	if table then
		return pairs(table)
	else
		return function() end
	end
end
